#!/usr/bin/env python3
#
############################################################################
#
# MODULE:      vector.py
# AUTHOR(S):   Anika Weinmann, Julia Haas
#
# PURPOSE:     functions for general vector processing
# COPYRIGHT:   (C) 2024 by mundialis GmbH & Co. KG and the GRASS
#              Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
############################################################################

import grass.script as grass


def patch_vectors(vector_list, output, rm_vectors=None):
    """Patch vector data from a list.

    Args:
        vector_list (list): List with vectors to patch
        output (str): Output map
        rm_vectors (list): List with vectors that should be removed

    """
    # patch several vectors (e.g. from parallel imports)
    if len(vector_list) > 1:
        grass.run_command(
            "v.patch",
            input=vector_list,
            output=output,
            flags="e",
            quiet=True,
        )
        if rm_vectors is not None:
            rm_vectors.extend(vector_list)
    else:
        grass.run_command(
            "g.rename",
            vector=f"{vector_list[0]},{output}",
            quiet=True,
        )



#!/usr/bin/env python3

############################################################################
#
# MODULE:       lib with validation related helper functions for GRASS GIS
#
# AUTHOR(S):    Lina Krisztian
#
# PURPOSE:      lib with validation related helper functions for GRASS GIS
#
# COPYRIGHT:	(C) 2023 by mundialis and the GRASS Development Team
#
# 		This program is free software under the GNU General Public
# 		License (>=v2). Read the file COPYING that comes with GRASS
# 		for details.
#
#############################################################################

import subprocess

import grass.script as grass


def get_gdalinfo_returncodes(input):
    """Return return codes from reading input file with gdalinfo."""
    gdalinfo_cmd = ["gdalinfo", "-mm", input]
    p_gdalinfo = subprocess.Popen(
        gdalinfo_cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    gdalinfo_err = p_gdalinfo.communicate()[1].decode("utf-8")
    gdalinfo_returncode = p_gdalinfo.returncode
    return gdalinfo_err, gdalinfo_returncode


def check_valid_rasterdata(input, strict=True):
    """Check if input is broken and returns grass.fatal() in this case."""
    gdalinfo_err, gdalinfo_returncode = get_gdalinfo_returncodes(input)
    if strict:
        # strict check: checks if gdalinfo contains any error
        if gdalinfo_err != "":
            grass.fatal(
                _(
                    f"<{input}> contains erroneous data.\n"
                    "NOTE: Might be harmless error messages. "
                    "Data might be still readable. "
                    "For a less strict check use: "
                    "check_valid_rasterdata(<input>,strict=False).",
                ),
            )
    # less strict check: fails only if bands can't be read
    elif gdalinfo_returncode != 0 or ("TIFFReadEncodedStrip" in gdalinfo_err):
        grass.fatal(_(f"<{input}> is a broken file"))



#!/usr/bin/env python3
#
############################################################################
#
# MODULE:      raster.py
# AUTHOR(S):   Anika Weinmann, Julia Haas
#
# PURPOSE:     functions for general raster processing
# COPYRIGHT:   (C) 2024 by mundialis GmbH & Co. KG and the GRASS
#              Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
############################################################################


import grass.script as grass


def adjust_raster_resolution(raster_name, output, res):
    """Resample or inpolate raster to given resolution. It is important that
    the region already has the right resolution.

    Args:
        raster_name (str): The name of the raster map which should be
                           resampled/interpolated
        output (str): The name for the resampled/interpolated raster map
        res (float): The resolution to which the raster should be resampled.

    """
    res_rast = float(
        grass.parse_command("r.info", map=raster_name, flags="g")["nsres"],
    )
    if res_rast > res:
        grass.run_command(
            "r.resamp.interp",
            input=raster_name,
            output=output,
            overwrite=True,
            quiet=True,
        )
    elif res_rast < res:
        grass.run_command(
            "r.resamp.stats",
            input=raster_name,
            output=output,
            method="median",
            quiet=True,
            overwrite=True,
        )
    else:
        rename_raster(raster_name, output)


def create_vrt(input_raster_list, output):
    """Create a VRT raster map out of input list, or renaming if only one
    raster is inside the list. If the input raster maps are inside other
    mapsets they will be copied to the current mapset before the VRT will be
    created.

    Args:
        input_raster_list (list): List with input raster maps
        output (str): Name of the output (vrt) raster map

    """
    # copy raster maps to current mapset
    for rast in input_raster_list:
        if "@" in rast:
            rast_wo_mapsetname = rast.split("@")[0]
            grass.run_command(
                "g.copy",
                raster=f"{rast},{rast_wo_mapsetname}",
            )
    input_raster_list = [val.split("@")[0] for val in input_raster_list]
    # buildvrt if required + renaming to output name
    if len(input_raster_list) > 1:
        grass.run_command("g.region", raster=input_raster_list)
        grass.run_command(
            "r.buildvrt",
            input=input_raster_list,
            output=output,
            quiet=True,
            overwrite=True,
        )
    else:
        grass.run_command(
            "g.rename",
            raster=f"{input_raster_list[0]},{output}",
            quiet=True,
            overwrite=True,
        )


def rename_raster(band_name_old, band_name_new):
    """Rename raster map.

    Args:
        band_name_old (str): Raster map name to rename
        band_name_new (str): The new name for the raster map

    """
    grass.run_command(
        "g.rename",
        raster=f"{band_name_old},{band_name_new}",
        quiet=True,
        overwrite=True,
    )



#!/usr/bin/env python3

############################################################################
#
# MODULE:       lib with cleanup related helper functions for GRASS GIS
#
# AUTHOR(S):    Anika Weinmann
#
# PURPOSE:      lib with cleanup related helper functions for GRASS GIS
#
# COPYRIGHT:	(C) 2023-2025 by mundialis and the GRASS Development Team
#
# 		This program is free software under the GNU General Public
# 		License (>=v2). Read the file COPYING that comes with GRASS
# 		for details.
#
#############################################################################

import os
import shutil
import gc

import grass.script as grass
from .location import get_location_size, switch_back_original_location


def general_cleanup(
    rm_rasters=[],
    rm_vectors=[],
    rm_files=[],
    rm_dirs=[],
    rm_groups=[],
    rm_groups_w_rasters=[],
    rm_groups_wo_rasters=[],  # kept for backwards compatibility
    rm_regions=[],
    rm_strds=[],
    rm_strds_w_rasters=[],
    rm_stvds=[],
    rm_stvds_w_vectors=[],
    orig_region=None,
    rm_mask=False,
    location_size=False,
):
    """General cleanup function."""
    grass.message(_("Cleaning up..."))
    nulldev = open(os.devnull, "w", encoding="utf-8")
    kwargs = {"flags": "f", "quiet": True, "stderr": nulldev}
    rm_groups.extend(rm_groups_wo_rasters)
    for rmg in rm_groups:
        if grass.find_file(name=rmg, element="group")["file"]:
            grass.run_command("g.remove", type="group", name=rmg, **kwargs)
    for rmg_wr in rm_groups_w_rasters:
        if grass.find_file(name=rmg_wr, element="group")["file"]:
            group_rasters = grass.parse_command(
                "i.group",
                flags="lg",
                group=rmg_wr,
            )
            rm_rasters.extend(group_rasters)
            grass.run_command("g.remove", type="group", name=rmg, **kwargs)
    for rmrast in rm_rasters:
        if grass.find_file(name=rmrast, element="raster")["file"]:
            grass.run_command("g.remove", type="raster", name=rmrast, **kwargs)
    for rmvect in rm_vectors:
        if grass.find_file(name=rmvect, element="vector")["file"]:
            grass.run_command("g.remove", type="vector", name=rmvect, **kwargs)
    for rmfile in rm_files:
        if os.path.isfile(rmfile):
            os.remove(rmfile)
    for rmdir in rm_dirs:
        if os.path.isdir(rmdir):
            shutil.rmtree(rmdir)
    if orig_region is not None:
        find_reg = grass.find_file(name=orig_region, element="windows")
        if find_reg.get("file"):
            grass.run_command("g.region", region=orig_region)
            grass.run_command(
                "g.remove",
                type="region",
                name=orig_region,
                **kwargs,
            )
    for rmreg in rm_regions:
        find_reg = grass.find_file(name=rmreg, element="windows")
        if find_reg.get("file"):
            grass.run_command("g.remove", type="region", name=rmreg, **kwargs)
    strds = grass.parse_command("t.list", type="strds", quiet=True)
    stvds = grass.parse_command("t.list", type="stvds", quiet=True)
    mapset = grass.gisenv()["MAPSET"]
    rm_strds_check = []
    for rm_s in rm_strds:
        if f"{rm_s}@{mapset}" in strds:
            rm_strds_check.append(rm_s)
    if len(rm_strds_check) > 0:
        grass.run_command(
            "t.remove",
            flags="rf",
            type="strds",
            input=rm_strds_check,
            quiet=True,
            stderr=nulldev,
        )
    rm_strds_w_rasters_check = []
    for rm_r_wr in rm_strds_w_rasters:
        if f"{rm_r_wr}@{mapset}" in strds:
            rm_strds_w_rasters_check.append(rm_r_wr)
    if len(rm_strds_w_rasters_check) > 0:
        grass.run_command(
            "t.remove",
            flags="fd",
            type="strds",
            input=rm_strds_w_rasters_check,
            quiet=True,
            stderr=nulldev,
        )
    rm_stvds_check = []
    for rm_v in rm_stvds:
        if f"{rm_v}@{mapset}" in stvds:
            rm_stvds_check.append(rm_v)
    if len(rm_stvds_check) > 1:
        grass.run_command(
            "t.remove",
            flags="rf",
            type="stvds",
            input=rm_stvds_check,
            quiet=True,
            stderr=nulldev,
        )
    rm_stvds_w_vectors_check = []
    for rm_v_wv in rm_stvds_w_vectors:
        if f"{rm_v_wv}@{mapset}" in stvds:
            rm_stvds_w_vectors_check.append(rm_v_wv)
    if len(rm_stvds_w_vectors_check) > 1:
        grass.run_command(
            "t.remove",
            flags="fd",
            type="stvds",
            input=rm_stvds_w_vectors_check,
            quiet=True,
            stderr=nulldev,
        )
    if rm_mask and grass.find_file(name="MASK", element="raster")["file"]:
        grass.run_command("r.mask", flags="r")

    # get location size
    if location_size:
        get_location_size()

    # Garbage Collector: release unreferenced memory
    gc.collect()


def rm_vects(vects):
    """Function to remove clean vector maps
    Args:
        vects (list): list of vector maps which should be removed.
    """
    nuldev = open(os.devnull, "w", encoding="utf-8")
    kwargs = {"flags": "f", "quiet": True, "stderr": nuldev}
    for rmv in vects:
        if grass.find_file(name=rmv, element="vector")["file"]:
            grass.run_command("g.remove", type="vector", name=rmv, **kwargs)


def reset_region(region):
    """Function to set the region to the given region
    Args:
        region (str): the name of the saved region which should be set and
                      deleted.
    """
    nulldev = open(os.devnull, "w", encoding="utf-8")
    kwargs = {"flags": "f", "quiet": True, "stderr": nulldev}
    if region and grass.find_file(name=region, element="windows")["file"]:
        grass.run_command("g.region", region=region)
        grass.run_command("g.remove", type="region", name=region, **kwargs)


def cleaning_tmp_location(original_gisrc, tmp_loc, gisdbase, tmp_gisrc):
    """Cleaning up things from temporary location.

    Args:
        original_gisrc (str): The path to the original GISRC file
        tmp_loc (str): The name of the temporary location
        gisdbase (str): The GISDBASE info
        tmp_gisrc (str): The path to the temporary GISRC file

    """
    # switch back to original gisrc
    if original_gisrc:
        switch_back_original_location(original_gisrc)
    # remove temporary location
    if tmp_loc:
        grass.try_rmdir(os.path.join(gisdbase, tmp_loc))
    if tmp_gisrc:
        grass.try_remove(tmp_gisrc)



#!/usr/bin/env python3

############################################################################
#
# MODULE:       lib with parallel related helper functions for GRASS GIS
#
# AUTHOR(S):    Anika Weinmann and Lina Krisztian
#
# PURPOSE:      lib with parallel related helper functions for GRASS GIS
#
# COPYRIGHT:	(C) 2023 by mundialis and the GRASS Development Team
#
# 		This program is free software under the GNU General Public
# 		License (>=v2). Read the file COPYING that comes with GRASS
# 		for details.
#
#############################################################################

import grass.script as grass
from grass.pygrass.modules import Module, ParallelModuleQueue

from .mapset import verify_mapsets


def check_parallel_errors(queue):
    """Check a parallel queue for errors in the worker modules by parsing
    the stderr output for errors. A GRASS fatal will be thrown in this case.
    To be used as except in a try/except block around a parallel GRASS
    processing queue.
    """
    for proc_num in range(queue.get_num_run_procs()):
        proc = queue.get(proc_num)
        if proc.returncode != 0:
            # save all stderr to a variable and pass it to a GRASS
            # exception
            errmsg = proc.outputs["stderr"].value.strip()
            grass.fatal(
                _(f"\nERROR processing <{proc.get_bash()}>: {errmsg}"),
            )


def check_parallel_warnings(queue):
    """Check a parallel queue for warnings in the worker modules by parsing
    the stderr output for warnings. A GRASS warning will be issued in this case.
    To be used after a successfully run processing queue.
    """
    for mod in queue.get_finished_modules():
        stderr = mod.outputs["stderr"].value.strip()
        if "WARN" in stderr:
            grass.warning(
                _(f"\nWARNING processing <{mod.get_bash()}>: {stderr}"),
            )


def run_module_parallel(
    module,
    module_kwargs,
    tile_list,
    nprocs,
    uid,
    parallel=True,
):
    """Running a module in parallel on a grid."""
    # save current mapset
    start_cur_mapset = grass.gisenv()["MAPSET"]

    queue = ParallelModuleQueue(nprocs=nprocs)
    mapset_names = []
    try:
        for num, tile in enumerate(tile_list):
            grass.message(_(f"Running {module} for tile {num} ..."))
            # create a unique mapset and vector subset name
            sid = f"{num}_{uid}"

            # Module
            new_mapset = f"tmp_{module.replace('.', '_')}_{sid}"
            mapset_names.append(new_mapset)
            module_kwargs["new_mapset"] = new_mapset
            module_kwargs["area"] = tile
            # memory=use_memory,
            if parallel:
                mod = Module(module, **module_kwargs, run_=False)
                # catch all GRASS outputs to stdout and stderr
                mod.stdout_ = grass.PIPE
                mod.stderr_ = grass.PIPE
                queue.put(mod)
            else:
                grass.run_command(module, **module_kwargs)
        # if parallel:
        queue.wait()
    except Exception:
        for proc_num in range(queue.get_num_run_procs()):
            proc = queue.get(proc_num)
            if proc.returncode != 0:
                # save all stderr to a variable and pass it to a GRASS
                # exception
                errmsg = proc.outputs["stderr"].value.strip()
                grass.fatal(
                    _(f"\nERROR by processing <{proc.get_bash()}>: {errmsg}"),
                )
    # print all logs of successfully run modules ordered by module as GRASS
    # message
    for proc in queue.get_finished_modules():
        msg = proc.outputs["stderr"].value.strip()
        grass.message(_(f"\nLog of {proc.get_bash()}:"))
        for msg_part in msg.split("\n"):
            if len(msg_part) > 0:
                grass.message(msg_part)

    # verify that switching the mapset worked
    location_path = verify_mapsets(start_cur_mapset)
    return mapset_names, location_path


def patching_vector_results(mapsets, output):
    """Patching vector results of different mapsets into one together."""
    grass.message(_(f"Patching vector {output} subsets ..."))
    if len(mapsets) > 1:
        subset_mapset = [f"{output}@{m}" for m in mapsets]
        grass.run_command(
            "v.patch",
            input=subset_mapset,
            output=output,
            flags="e",
            overwrite=True,
            quiet=True,
        )
    else:
        grass.run_command(
            "g.copy",
            vector=f"{output}@{mapsets[0]},{output}",
            overwrite=True,
        )


def patching_raster_results(mapsets, output):
    """Patching raster results of different mapsets into one together."""
    grass.message(_(f"Patching raster {output} subsets ..."))
    if len(mapsets) > 1:
        subset_mapset = [f"{output}@{m}" for m in mapsets]
        grass.run_command(
            "r.patch",
            input=subset_mapset,
            output=output,
            overwrite=True,
            quiet=True,
        )
    else:
        grass.run_command(
            "g.copy",
            raster=f"{output}@{mapsets[0]},{output}",
            overwrite=True,
        )



#!/usr/bin/env python3

############################################################################
#
# MODULE:       lib with test related helper functions for GRASS GIS
#
# AUTHOR(S):    Anika Weinmann
#
# PURPOSE:      lib with test related helper functions for GRASS GIS
#
# COPYRIGHT:	(C) 2024 by mundialis and the GRASS Development Team
#
# 		This program is free software under the GNU General Public
# 		License (>=v2). Read the file COPYING that comes with GRASS
# 		for details.
#
#############################################################################

import grass.script as grass


def get_number_of_grass_elements():
    """Get the number of grass elements like raster, vector and regions.

    Returns:
        num_rast (int): The number of raster maps in current mapset
        num_vect (int): The number of vector maps in current mapset
        num_gr (int): The number of groups in current mapset
        num_reg (int): The number of regions in current mapset
        num_mapsets (int): The number of mapsets in current location

    """
    rast_list = grass.parse_command("g.list", type="raster")
    vect_list = grass.parse_command("g.list", type="vector")
    gr_list = grass.parse_command("g.list", type="group")
    reg_list = grass.parse_command("g.list", type="region")
    mapset_list = grass.parse_command("g.mapsets", flags="l", quiet=True)
    num_rast = len(rast_list)
    num_vect = len(vect_list)
    num_gr = len(gr_list)
    num_reg = len(reg_list)
    num_mapsets = len(mapset_list)
    return num_rast, num_vect, num_gr, num_reg, num_mapsets


def check_number_of_grass_elements(
    ref_num_rast,
    ref_num_vect,
    ref_num_gr,
    ref_num_reg,
    ref_num_mapsets,
):
    """Check the number of grass elements.

    Args:
        ref_num_rast (int): The reference number of raster maps
        ref_num_vect (int): The reference number of vector maps
        ref_num_gr (int): The reference number of groups
        ref_num_reg (int): The reference number of regions
        ref_num_mapsets (int): The reference number of mapsets

    """
    n_rast, n_vect, n_gr, n_reg, n_mapsets = get_number_of_grass_elements()
    for num, ref, data_type in [
        (n_rast, ref_num_rast, "rasters"),
        (n_vect, ref_num_vect, "vectors"),
        (n_gr, ref_num_gr, "groups"),
        (n_reg, ref_num_reg, "regions"),
        (n_mapsets, ref_num_mapsets, "mapsets"),
    ]:
        assert (
            num == ref
        ), f"The number of {data_type} is {num} but should be {ref}"



#!/usr/bin/env python3
#
############################################################################
#
# MODULE:      data_import.py
# AUTHOR(S):   Anika Weinmann, Julia Haas
#
# PURPOSE:     functions for the local data import
# COPYRIGHT:   (C) 2024 by mundialis GmbH & Co. KG and the GRASS
#              Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
############################################################################

import glob
import os
import subprocess
from subprocess import PIPE
import wget

import grass.script as grass

from .cleanup import rm_vects
from .general import communicate_grass_command
from .raster import adjust_raster_resolution, rename_raster
from .vector import patch_vectors


def download_and_import_tindex(tindex_url, output, download_dir):
    """Download and import tile index from url.

    Args:
        tindex_url (str): URL of tile index
        output (str): The output name for the tile index
        download_dir (str): The directory where the data should be downloaded

    """
    cur_dir = os.getcwd()
    zip_name = os.path.basename(tindex_url)
    tindex_gpkg = zip_name.replace(".gz", "")
    try:
        os.chdir(download_dir)
        # download data
        wget.download(tindex_url, zip_name, bar=None)

        # unzip tindex
        os.system(f"gunzip {zip_name}")

        # import data
        grass.run_command(
            "v.import",
            input=tindex_gpkg,
            output=output,
            extent="region",
            overwrite=True,
            quiet=True,
        )
    finally:
        for rm_file in [zip_name, tindex_gpkg]:
            if os.path.isfile(rm_file):
                os.remove(rm_file)
        os.chdir(cur_dir)


def get_list_of_tindex_locations(tindex, aoi=None):
    """Select the locations of the tindex which overlap with the AOI or the
    current region.

    Args:
        tindex (str): Name of the tindex vector map
        aoi (str): Name of the AOI vector map
    Returns:
        (list): List with locations which overlap with the AOI or the current
                region

    """
    tindex_clipped = f"clipped_tindex_vect_{grass.tempname(8)}"
    try:
        v_clip_kwargs = {
            "input": tindex,
            "output": tindex_clipped,
            "flags": "",
            "quiet": True,
        }
        if aoi:
            v_clip_kwargs["clip"] = aoi
            v_clip_kwargs["flags"] += "d"
        else:
            v_clip_kwargs["flags"] += "r"
        grass.run_command("v.clip", **v_clip_kwargs)
        # grass.vector_db_select(tindex_clipped, columns="location")
        tiles = [
            val[0]
            for val in grass.vector_db_select(
                tindex_clipped,
                columns="location",
            )["values"].values()
        ]
    finally:
        rm_vects([tindex_clipped])
    return tiles


def import_local_raster_data(
    aoi,
    basename,
    local_data_dir,
    native_res_flag,
    all_raster,
    rm_rasters,
    band_dict=None,
):
    """Import local raster data. Where a VRT or TIFs are given in the directory
    of "local_data_dir".

    Args:
        aoi (str): Vector map with area of interest
        basename (str): Basename for imported rasters
        local_data_dir (str): Path to local data directory with VRT or TIF
                              files inside
        native_res_flag (bool): True if native data resolution should be used
        all_raster (list/dict): Empty list/dictionary where the imported rasters
                                will be appended
        rm_rasters (list): List with rasters which should be removed
        band_dict (dict): Dictionary with band number and names, if none only
                          one band should be in the files which should be
                          imported; e.g. for DOP import band_dict = {
                              1: "red", 2: "green", 3: "blue", 4: "nir"
                          }

    Returns:
        imported_local_data (bool): True if local data imported, otherwise
                                    False

    """
    grass.message(_("Importing local raster data..."))
    imported_local_data = False
    if band_dict is None:
        band_dict = {"": ""}
    # get files (VRT if available otherwise TIF)
    raster_files = glob.glob(
        os.path.join(local_data_dir, "**", "*.vrt"),
        recursive=True,
    )
    if not raster_files:
        raster_files = glob.glob(
            os.path.join(local_data_dir, "**", "*.tif"),
            recursive=True,
        )

    # get current region
    cur_reg = grass.region()
    ns_res = cur_reg["nsres"]
    ew_res = cur_reg["ewres"]
    # import data for AOI
    # TODO parallelize local data import for multi TIFs/VRTs
    for i, raster_file in enumerate(raster_files):
        # set aoi if it is given with current resolution
        if aoi and aoi != "":
            grass.run_command(
                "g.region",
                vector=aoi,
                nsres=ns_res,
                ewres=ew_res,
                flags="a",
                quiet=True,
            )
            # grow region because of interpolation
            grass.run_command("g.region", grow=1, quiet=True)
        # Import data
        name = f"{basename}_{i}"
        kwargs = {
            "input": raster_file,
            "output": name,
            "extent": "region",
            "quiet": True,
            "overwrite": True,
        }
        # resolution settings: -r native resolution; otherwise from region
        if not native_res_flag:
            kwargs["resolution"] = "region"
        r_import = communicate_grass_command("r.import", **kwargs)
        err_m1 = "Input raster does not overlap current computational region."
        err_m2 = "already exists and will be overwritten"
        if err_m1 in r_import[1].decode():
            continue
        if err_m2 in r_import[1].decode():
            pass
        elif r_import[1].decode() != "":
            grass.fatal(_(r_import[1].decode()))

        # resample / interpolate data
        for band_num, band in band_dict.items():
            band_name_old = f"{name}.{band_num}" if band_num != "" else name
            band_name_new = f"{name}.{band}" if band_num != "" else name
            rm_rasters.append(band_name_old)
            # check resolution and resample if needed
            if not native_res_flag:
                adjust_raster_resolution(band_name_old, band_name_new, ns_res)
            else:
                rename_raster(band_name_old, band_name_new)
            if isinstance(all_raster, dict):
                all_raster[band].append(band_name_new)
            else:
                all_raster.append(band_name_new)

    # check if all bands have at least one input raster
    if isinstance(all_raster, dict):
        if all([len(val) > 0 for band, val in all_raster.items()]):
            imported_local_data = True
    elif len(all_raster) > 0:
        imported_local_data = True

    return imported_local_data


def get_xyz_file_infos(xyz_file, separator="space"):
    """Get the infos of a XYZ file to resolution, bounding box and pixelcenter.

    Args:
        xyz_file (str): XYZ file path to import
        separator (str): Separator of XYZ file; default is "space"
    Returns:
        res (float): Resolution of the XYZ file
        xyz_reg (dict): Dictionary with region of the XYZ file
        shift_needed (bool): Boolean if the XYZ file hat to be shifted

    """
    gdalinfo_cmd = ["gdalinfo", xyz_file]
    process = subprocess.Popen(gdalinfo_cmd, stdout=PIPE, stderr=PIPE)
    stdout = process.communicate()[0].decode("utf-8").strip()
    # output vom ogrinfo:
    # z.B.: Pixel Size = (1.000000000000000,1.000000000000000)
    res = float(stdout.split("Pixel Size = (")[1].split(",")[0])
    # get bbox
    bbox_x1 = float(
        stdout.split("Upper Left")[1].replace("(", "").split(",")[0].strip(),
    )
    bbox_x2 = float(
        stdout.split("Upper Right")[1].replace("(", "").split(",")[0].strip(),
    )
    bbox_y1 = float(
        stdout.split("Upper Left")[1].split(",")[1].split(")")[0].strip(),
    )
    bbox_y2 = float(
        stdout.split("Lower Left")[1].split(",")[1].split(")")[0].strip(),
    )
    # check if shift is needed
    # The shift is only needed if the bbox does not contain the pixel centers
    # (in the bbox the half of the resolution could be seen)
    shift_needed = True
    half_res = res / 2.0
    if (
        bbox_x1 % res != half_res
        and bbox_x2 % res != half_res
        and bbox_y1 % res != half_res
        and bbox_y2 % res != half_res
    ):
        shift_needed = False
    # get region to xyz file
    xyz_reg_str = grass.read_command(
        "r.in.xyz",
        output="dummy",
        input=xyz_file,
        flags="sg",
        separator=separator,
    )
    xyz_reg = {
        item.split("=")[0]: float(item.split("=")[1])
        for item in xyz_reg_str.strip().split(" ")
    }
    xyz_reg["n"] += half_res
    xyz_reg["s"] -= half_res
    xyz_reg["w"] -= half_res
    xyz_reg["e"] += half_res
    return res, xyz_reg, shift_needed


def import_single_local_xyz_file(
    xyz_file,
    output,
    use_cur_reg=False,
    separator="space",
):
    """Import single XYZ file.

    Args:
        xyz_file (str): XYZ file path to import
        output (str): Output raster file name
        use_cur_reg (bool): If True the XYZ file will only be imported if it
                            overlaps with the current region, otherwise it
                            will not be imported
        separator (str): Separator of XYZ file; default is "space"
    Returns:
        output (str): If the output is imported, otherwise return None

    """
    res, xyz_reg, shift_needed = get_xyz_file_infos(
        xyz_file,
        separator=separator,
    )
    # check if aoi overlaps
    if use_cur_reg:
        cur_reg = grass.region()
        if (
            cur_reg["e"] < xyz_reg["w"]
            or xyz_reg["e"] < cur_reg["w"]
            or cur_reg["n"] < xyz_reg["s"]
            or xyz_reg["n"] < cur_reg["s"]
        ):
            return None
    # set region
    grass.run_command(
        "g.region",
        n=xyz_reg["n"],
        s=xyz_reg["s"],
        w=xyz_reg["w"],
        e=xyz_reg["e"],
        res=res,
    )
    if use_cur_reg:
        while (cur_reg["n"] + res) < xyz_reg["n"]:
            grass.run_command("g.region", n=f"n-{res}")
            xyz_reg["n"] -= res
        while (cur_reg["s"] - res) > xyz_reg["s"]:
            grass.run_command("g.region", s=f"s+{res}")
            xyz_reg["s"] += res
        while (cur_reg["e"] + res) < xyz_reg["e"]:
            grass.run_command("g.region", e=f"e-{res}")
            xyz_reg["e"] -= res
        while (cur_reg["w"] - res) > xyz_reg["w"]:
            grass.run_command("g.region", w=f"w+{res}")
            xyz_reg["w"] += res
    # import data
    grass.run_command(
        "r.in.xyz",
        input=xyz_file,
        output=output,
        method="mean",
        separator=separator,
        quiet=True,
        overwrite=True,
    )
    # shift data if needed
    if shift_needed:
        grass.run_command(
            "g.region",
            n=f"n+{res / 2.}",
            s=f"s+{res / 2.}",
            w=f"w+{res / 2.}",
            e=f"e+{res / 2.}",
            res=res,
        )
        grass.run_command("r.region", map=output, flags="c")
    return output


def import_local_xyz_files(
    aoi,
    basename,
    local_data_dir,
    all_raster,
):
    """Import local XYZ raster data.  XYZ files which are inside the
    directory of "local_data_dir" will be imported for the AOI.

    Args:
        aoi (str): Vector map with area of interest
        basename (str): Basename for imported rasters
        local_data_dir (str): Path to local data directory with XYZ files
        all_raster (list/dict): empty list/dictionary where the imported rasters
                                will be appended
    Returns:
        imported_local_data (bool): True if local data imported, otherwise False

    """
    grass.message(_("Importing local XYZ data..."))
    imported_local_data = False

    # get XYZ files
    xyz_files = glob.glob(
        os.path.join(local_data_dir, "**", "*.xyz"),
        recursive=True,
    )

    # import data for AOI
    # TODO parallelize local data import
    # get current region
    cur_reg = grass.region()
    ns_res = cur_reg["nsres"]
    ew_res = cur_reg["ewres"]
    for i, xyz_file in enumerate(xyz_files):
        # set aoi if it is given with current resolution
        if aoi and aoi != "":
            grass.run_command(
                "g.region",
                vector=aoi,
                nsres=ns_res,
                ewres=ew_res,
                flags="a",
                quiet=True,
            )
            # grow region because of interpolation
            grass.run_command("g.region", grow=1, quiet=True)
        # Import data
        name = f"{basename}_{i}"
        name = import_single_local_xyz_file(xyz_file, name, True)
        if name:
            all_raster.append(name)
            grass.message(
                _(f"XYZ file <{os.path.basename(xyz_file)}> imported."),
            )
    # check if raster were imported
    if len(all_raster) > 0:
        imported_local_data = True
    return imported_local_data


def import_local_vector_data(aoi_map, local_data_dir, rm_vectors, output):
    """Import vector data from local file path.

    Args:
        aoi_map (str): Name of vector map defining AOI
        local_data_dir (str): Path to local data
        rm_vectors (list): List with vectors that should be removed
        output (str): Output map

    Returns:
        imported_local_data (bool): True if local data imported, otherwise False

    """
    imported_local_data = False

    # get files (GPKG or SHP)
    files = glob.glob(
        os.path.join(local_data_dir, "**", "*.gpkg"),
        recursive=True,
    )
    shp_files = glob.glob(
        os.path.join(local_data_dir, "**", "*.shp"),
        recursive=True,
    )
    files.extend(shp_files)

    # import data for AOI
    imported_list = []
    for i, file in enumerate(files):
        if aoi_map:
            grass.run_command(
                "g.region",
                vector=aoi_map,
                quiet=True,
            )
        grass.run_command(
            "v.import",
            input=file,
            output=f"{output}_{i}",
            extent="region",
            quiet=True,
        )
        imported_list.append(f"{output}_{i}")

    # patch outputs
    patch_vectors(imported_list, output, rm_vectors)

    # check if result has at least one feature
    map_info = grass.parse_command(
        "v.info",
        map=output,
        flags="gt",
    )
    if int(map_info["centroids"]) > 0:
        imported_local_data = True

    return imported_local_data



#!/usr/bin/env python3
#
############################################################################
#
# MODULE:      federal_state.py
# AUTHOR(S):   Anika Weinmann
#
# PURPOSE:     open-geodata-germany: infos and functions to federale states
# COPYRIGHT:   (C) 2024 by mundialis GmbH & Co. KG and the GRASS
#              Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
############################################################################

import os
import requests
import grass.script as grass

FS_ABBREVIATION = {
    "Baden-Württemberg": "BW",
    "BW": "BW",
    "Bayern": "BY",
    "BY": "BY",
    "Berlin": "BE",
    "BE": "BE",
    "Brandenburg": "BB",
    "BB": "BB",
    "Bremen": "HB",
    "HB": "HB",
    "Hamburg": "HH",
    "HH": "HH",
    "Hessen": "HE",
    "HE": "HE",
    "Mecklenburg-Vorpommern": "MV",
    "MV": "MV",
    "Niedersachsen": "NI",
    "NI": "NI",
    "Nordrhein-Westfalen": "NW",
    "NW": "NW",
    "Rheinland-Pfalz": "RP",
    "RP": "RP",
    "Saarland": "SL",
    "SL": "SL",
    "Sachsen": "SN",
    "SN": "SN",
    "Sachsen-Anhalt": "ST",
    "ST": "ST",
    "Schleswig-Holstein": "SH",
    "SH": "SH",
    "Thüringen": "TH",
    "TH": "TH",
}


def import_administrative_boundaries(output, aoi=None, level="KRS"):
    """Import administrative boundaries for AOI/region.

    Args:
        output (str): The name for the output vector map with the imported
                      administrative boundaries
        aoi (str): The name of the area of interest for which the
                   administrative boundaries should be imported; if aoi is set
                   to None the boundaries will be imported for the current
                   region
        level (str): The level of the administrative boundaries;
                     STA - Staat
                     LAN - Länder
                     RBZ - Regierungsbezirke
                     KRS - Kreise
                     VWG - Verwaltungsgemeinschaften
                     GEM - Gemeinden

    """
    # save current region and set region to AOI
    if aoi:
        reg = f"tmp_region_{grass.tempname(8)}"
        grass.run_command("g.region", save=reg)
        grass.run_command("g.region", vector=aoi, quiet=True)

    # url of administrative boundaries
    url = (
        "https://daten.gdz.bkg.bund.de/produkte/vg/vg5000_0101/"
        "aktuell/vg5000_01-01.utm32s.shape.ebenen.zip"
    )
    # file of administrative boundaries in zip
    filename = os.path.join(
        "vg5000_01-01.utm32s.shape.ebenen",
        "vg5000_ebenen_0101",
        f"VG5000_{level}.shp",
    )
    try:
        # check if URL is reachable
        response = requests.get(url)
        if response.status_code != 200:
            grass.fatal(
                "The data import of the administrative boundaries are "
                "currently not available.",
            )

        # download and import administrative boundaries
        vsi_url = f"/vsizip/vsicurl/{url}/{filename}"
        grass.run_command(
            "v.import",
            input=vsi_url,
            output=output,
            extent="region",
            quiet=True,
        )
    finally:
        if aoi:
            grass.run_command("g.region", region=reg)


def get_federal_states(federal_state, federal_state_file):
    """Get federal state and federal state file module parameters and return
    list with federal state abbreviations.

    Args:
        federal_state (str): Federal state module parameter
        federal_state_file (str): Federal state file module parameter
    Returns:
        (list): List with federal state abbreviations

    """
    if federal_state_file:
        if not os.path.isfile(federal_state_file):
            grass.fatal(
                _(
                    "Federal state file is given, but file "
                    f"<{federal_state_file}> does not exist.",
                ),
            )
        with open(federal_state_file, encoding="utf-8") as fs_file:
            fs_list_str = fs_file.read().strip()
            if fs_list_str == "":
                grass.fatal(
                    _(
                        "Federal state in <federal_state_file> is empty "
                        "string!",
                    ),
                )
    elif federal_state:
        fs_list_str = federal_state.strip()
    else:
        grass.fatal(
            _(
                "Neither <federal_state> nor <federal_state_file> are given. "
                "Please set one of the two.",
            ),
        )
    fs_list = []
    for fs in fs_list_str.split(","):
        fs = fs.strip()
        if fs not in FS_ABBREVIATION:
            grass.fatal(_(f"Non valid name of federal state: {fs}"))
        fs_list.append(FS_ABBREVIATION[fs])
    return fs_list



#!/usr/bin/env python3
#
############################################################################
#
# MODULE:      download_data.py
# AUTHOR(S):   Anika Weinmann
#
# PURPOSE:     open-geodata-germany: functions for the download
# COPYRIGHT:   (C) 2024 by mundialis GmbH & Co. KG and the GRASS
#              Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
############################################################################

import fileinput
from multiprocessing.pool import ThreadPool
import os
from zipfile import ZipFile

import grass.script as grass
import requests
import zipfile_deflate64


def check_download_dir(download_dir):
    """Check if download directory is set. If yes, check if folder exists or
    create it. If not set, a temporary directory will be used.

    Args:
        download_dir (str): Download directory module parameter
    Returns:
        (str): Path to download directory

    """
    if not download_dir:
        download_dir = grass.tempdir()
    elif not os.path.isdir(download_dir):
        grass.message(
            _(
                f"Download folder {download_dir} does not exist and will "
                "be created.",
            ),
        )
        os.makedirs(download_dir)
    elif os.path.exists(download_dir) and os.listdir(download_dir):
        grass.warning(
            _(
                f"Download folder {download_dir} exists and is not empty. "
                "Folder will NOT be deleted.",
            ),
        )
    grass.message(f"Download directory: {download_dir}")
    return download_dir


def url_response(url):
    """URL response function which is used by download_data_using_threadpool.

    Args:
        url (str): Data download url
    Return:
        url (str): Return the url for printing

    """
    filename = os.path.basename(url)
    response = requests.get(url, stream=True)
    with open(str(filename), "wb") as f:
        f.writelines(response)
    return url


def download_data_using_threadpool(urls, download_dir, nprocs):
    """Download data from urls via ThreadPool.

    Args:
        urls (list): List with data download urls
        download_dir (str): Path to directory where the data should be
                            downloaded to
        nprocs (int): The number of worker threads to use; If processes is None
                      then the number returned by os.cpu_count() is used.

    """
    cur_dir = os.getcwd()
    try:
        os.chdir(download_dir)
        pool = ThreadPool(nprocs)
        results = pool.imap_unordered(url_response, urls)
        grass.message(_("Downloading data of url:"))
        for result in results:
            print(result)
    finally:
        os.chdir(cur_dir)


def extract_compressed_files(file_names, download_dir):
    """Extract compressed files to download directory.

    Args:
        file_names (list): List with compressed e.g. zip file names which
                           are stored in the download_dir
        download_dir (str): Path to directory where the data should be
                            downloaded
    Returns:
        extracted_files (list): List with extracted files

    """
    extracted_files = []
    for file_name in file_names:
        file = os.path.join(download_dir, file_name)
        with ZipFile(file, "r") as zipobj:
            zip_content = zipobj.namelist()
            # Extract all the contents of zip file in current directory
            zipobj.extractall(download_dir)
            extracted_files.extend(zip_content)
    return extracted_files


def extract_compressed_files_deflate64(file_names, download_dir):
    """Extract compressed files to download directory using the
    zipfile_deflate64 library.

    Args:
        file_names (list): List with compressed e.g. zip file names which
                           are stored in the download_dir
        download_dir (str): Path to directory where the data should be
                            downloaded
    Returns:
        extracted_files (list): List with extracted files

    """
    extracted_files = []
    for file_name in file_names:
        file = os.path.join(download_dir, file_name)
        with zipfile_deflate64.ZipFile(file, "r") as zipobj:
            zip_content = zipobj.namelist()
            # Extract all the contents of zip file in current directory
            zipobj.extractall(download_dir)
            extracted_files.extend(zip_content)
    return extracted_files


def fix_corrupted_data(file):
    """Fix corrupted XYZ/TXT data file e.g. for Berlin DOMs.

    Args:
        file (str): XYZ or TXT data file with corrupted data

    """
    # remove corrupted data from TXT DOM files
    if not os.path.exists(f"{file}.bak"):
        with fileinput.FileInput(
            file,
            inplace=True,
            backup=".bak",
        ) as file_object:
            for line in file_object:
                # two times replace of white spaces, since some lines contain
                # 3 spaces
                print(
                    line.replace("  ", " ")
                    .replace("  ", " ")
                    .replace("\t", " "),
                    end="",
                )



#!/usr/bin/env python3

############################################################################
#
# MODULE:       lib with general related helper functions for GRASS GIS
#
# AUTHOR(S):    Anika Weinmann
#
# PURPOSE:      lib with general related helper functions for GRASS GIS
#
# COPYRIGHT:	(C) 2023 by mundialis and the GRASS Development Team
#
# 		This program is free software under the GNU General Public
# 		License (>=v2). Read the file COPYING that comes with GRASS
# 		for details.
#
#############################################################################

import multiprocessing as mp
import subprocess
import psutil

import grass.script as grass


def set_nprocs(nprocs):
    """Set nprocs to value if it is -2, otherwise check value."""
    if isinstance(nprocs, str):
        nprocs = int(nprocs)
    if nprocs == -2:
        return mp.cpu_count() - 1 if mp.cpu_count() > 1 else 1
    nprocs_real = mp.cpu_count()
    if nprocs > nprocs_real:
        grass.warning(
            f"Using {nprocs} parallel processes but only "
            f"{nprocs_real} CPUs available.",
        )
    return nprocs


def communicate_grass_command(*args, **kwargs):
    """Return stdout and stderr from executed GRASS command."""
    kwargs["stdout"] = grass.PIPE
    kwargs["stderr"] = grass.PIPE
    grass_ps = grass.start_command(*args, **kwargs)
    return grass_ps.communicate()


def check_grass_version(comp_version=(8, 0, 0)):
    """Returns boolean, if current GRASS version is >= some compare version."""
    cur_version = tuple(
        [
            int(x.replace("dev", "")) if x != "dev" else 0
            for x in grass.version()["version"].split(".")
        ],
    )
    return cur_version >= comp_version


def log_memory(grassenv=None):
    """Log memory usage."""
    if not grassenv:
        grassenv = grass.gisenv()
    cmd = subprocess.Popen(
        ["df", "-h", grassenv["GISDBASE"]],
        stdout=subprocess.PIPE,
    )
    grass.message(
        _(
            (
                "\nDisk usage of GRASS GIS database:\n",
                f"{cmd.communicate()[0].decode('utf-8').rstrip()}\n",
            ),
        ),
    )

    grass.message(_(f"\nmemory: \n{psutil.virtual_memory()!s}"))
    grass.message(_(f"\nswap memory: \n{psutil.swap_memory()!s}"))
    # ulimit -a
    cmd = subprocess.Popen(["ulimit", "-a"], stdout=subprocess.PIPE)
    grass.message(
        _(f"\nulimit -a: \n{cmd.communicate()[0].decode('utf-8').rstrip()}"),
    )


def get_free_ram(unit, percent=100):
    """The function gives the amount of the percentages of the available
    RAM memory and free swap space.

    Args:
        unit(string): 'GB' or 'MB'
        percent(int): number of percent which should be used of the available
                      RAM memory and free swap space
                      default 100%
    Returns:
        memory_MB_percent/memory_GB_percent(int): percent of the the available
                                                  memory and free swap in MB or
                                                  GB.

    """
    # use psutil cause of alpine busybox free version for RAM/SWAP usage
    mem_available = psutil.virtual_memory().available
    swap_free = psutil.swap_memory().free
    memory_gb = (mem_available + swap_free) / 1024.0**3
    memory_mb = (mem_available + swap_free) / 1024.0**2

    if unit == "MB":
        memory_mb_percent = memory_mb * percent / 100.0
        return round(memory_mb_percent)
    if unit == "GB":
        memory_gb_percent = memory_gb * percent / 100.0
        return round(memory_gb_percent)
    grass.fatal(f"Memory unit {unit} not supported")
    return None


def test_memory(memory_string):
    """Test if desired memory is available. In case RAM is smaller than
    desired memory, use free RAM instead of desired memory value.

    Args:
        memory_string(string): string from standard memory input option
    Returns:
        free_ram(int): free RAM to use
        memory(int): available memory to use.

    """
    # check memory
    memory = int(memory_string)
    free_ram = get_free_ram("MB", 100)
    if free_ram < memory:
        grass.warning(
            _(f"Using {memory} MB but only {free_ram} MB RAM available."),
        )
        grass.warning(_(f"Set used memory to {free_ram} MB."))
        return free_ram
    return memory


def check_installed_addon(addon, url="..."):
    """Check if addon is already installed and raise error if not.

    Args:
        addon(string): Addon to check if it is installed
        url(string): Path to addon.

    """
    if not grass.find_program(addon, "--help"):
        msg = (
            f"The '{addon}' module was not found, install  it first:\n"
            f"g.extension {addon} url={url}"
        )
        grass.fatal(_(msg))
